/*
 * Glavni fajl samog modula, sadrzi funkcije za ucitavanje i izbacivanje
 * modula, funkcije koje obradjuju operacije nad fajlom uredjaja kao i
 * funkcije koje vrse samu kontrolu PC zvucnika
 */
 
#include <linux/init.h> 	/*module_init(), module_exit(), __init, __exit*/
#include <linux/module.h>	/*za sve module*/
#include <linux/kdev_t.h>	/*dev_t*/
#include <linux/errno.h>	/*makroi za greske*/
#include <linux/moduleparam.h>	/*module_param()*/
#include <linux/sched.h>	/*current*/
#include <linux/fs.h>		/*fops, alloc_chrdev_region(), unregister_chrdev_region(),*/
#include <linux/cdev.h>		/*cdev, cdev_init, cdev_alloc, cdev_add, cdev_del*/
#include <linux/vmalloc.h>		/*vmalloc()*/
#include <linux/string.h>	/*strcpy(), strcat(), strlen()...*/
#include <asm/uaccess.h>	/*copy_to_user(), copy_from_user()*/
#include <linux/semaphore.h>	/*struct semaphore, sema_init(), dow_interruptible(), up() */
#include <linux/ioctl.h>	/*request_region()*/
#include <linux/delay.h> 	/*msleep_interruptible()*/
#include <linux/kthread.h>	/*task struct, kthread_create(), wake_up_process()*/
#include <linux/completion.h>	/*struct completion, init_completion(), complete()*/
#include <linux/capability.h>	/*capable()	*/
#include <linux/ioport.h>	/*inb(), outb()*/
#include <asm/io.h>		/*-||-*/
#include <linux/spinlock.h>  	/*spin_lock_irqsave(), spin_lock_irqrestore(), spinlock_t*/
#include <linux/version.h>
#include <linux/i8253.h>

/* 
 * smp_mb() je pomeren u drugi heder od verzije 3.3
 */
#if LINUX_VERSION_CODE<=KERNEL_VERSION(3, 3, 0)
	#include <asm/system.h>			/*smp_mb()*/
#else
	#include <asm/barrier.h>
#endif

/*
 * Metapodaci o modulu
 */
MODULE_DESCRIPTION("A module that translates text written into /dev/telegraf\
and plays it over the PC speaker");
MODULE_AUTHOR("Nikola Pavlović <nikola825@gmail.com>");
MODULE_VERSION("0.1");
MODULE_LICENSE("GPL");

/*
 * Slede konstante koje se koriste u modulu
 */
#define DEFAULTFREQ 500		// podrazumevana frekvencija zvucnika
#define DEVNAME "telegraf"	// ime koje se vidi u /proc/devices
#define DEVCOUNT 1			// broj uredjaja koji se registruje

//poruka koja se vraca ako neko pokusa da procita /dev/telegraf	
#define READ_RESPONSE "This file isn't ment to be read\ntype modinfo \
pc_speaker_telegraf for more information\n"

#define PLAY_THREAD_NAME "playthread"	//ime niti koja pusta zvuk
#define TIMERFREQ 1193180				//frekvencija i8253 oscilatora
#define DEFAULT_DOTLENGTH 100			//period Morzeove tacke u ms

//kontrolni kod za azuriranje frekvencije i8253 PIT2 tajmera
#define CONTROLL_CODE_UPDATE_COUNTDOWN 0xb6	

#define PIT2_CONTROLL_PORT 0x43		//kontrolni port za podesavanje
#define COUNTDOWN_UPDATE_PORT 0x42	//kontrolni port za setovanje frekvencije

//kontrolni port za paljenje/gasenje zvucnika i tajmera
#define ON_OFF_PORT 0x61		

#define IOCTLCODE_SETFREQ 1		// kodovi operacija za ioctl poziv
#define IOCTLCODE_SETDOTLENGTH 3

// pomocni makroi za pauziranje modula u duzini tacke i crte
#define DOTSLEEP msleep_interruptible(dotlength)	
#define LETTERSLEEP msleep_interruptible(dotlength*2)

/*
 * Prototipi funkcija koje se koriste u modulu
 */
static int  __init pc_speaker_telegraf_init(void);
static void __exit pc_speaker_telegraf_exit(void);	
static ssize_t telegraf_read(struct file *read_file, char __user *readbuffer, size_t read_size, loff_t *location); 
static ssize_t telegraf_write(struct file *write_file, const char __user *writebuffer, size_t write_size, loff_t *location); 
static long telegraf_ioctl(struct file *telegraf_file, unsigned int ioctl_command, unsigned long ioctl_args); 
static void playchar(char letter);
static int playstring(void *str);
static void playdot(void);
static void playdash(void);

/*
 * Struktura koja definise funkcije koje obradjuju sistemske pozive
 * read write i ioctl za fajl /dev/telegraf
 */

static struct file_operations telegraf_fops=
{
	.owner=				THIS_MODULE,
	.read=				telegraf_read,
	.write=				telegraf_write,
	.unlocked_ioctl=	telegraf_ioctl,
};


static enum plstat{FORCE_STOP, GO_ON} play_status=GO_ON;		
static struct cdev telegraf_cdev;		
static dev_t telegraf_devnum;			
static int freq=DEFAULTFREQ;			
static int countdown=TIMERFREQ/DEFAULTFREQ;	
static int dotlength=DEFAULT_DOTLENGTH;		
static struct semaphore write_semafor, ioctl_semafor, read_semafor;	
static struct completion write_completion;
static unsigned int highbyte=(TIMERFREQ/DEFAULTFREQ)/256;
static unsigned int lowbyte=(TIMERFREQ/DEFAULTFREQ)-((TIMERFREQ/DEFAULTFREQ)/256);
static bool dev_ok=false, cdev_ok=false;

/*
 * Makroi koji definisu parametre koji se mogu proslediti insmod-u
 * parametri se definisu staticki u vreme kompajliranja
 */
module_param(freq, int, S_IRUGO);
module_param(dotlength, int, S_IRUGO);

/*
 * Makroi koji definisu funkcije koje se pozivaju pri ucitavanju i izbacivanju
 * modula. Ove dve funkcije moraju biti poznate staticki u vreme kompajliranja
 */
module_init(pc_speaker_telegraf_init);
module_exit(pc_speaker_telegraf_exit);


static int  __init pc_speaker_telegraf_init(void)
/*
 * Funkcija koja se poziva pri ucitavanju modula
 * Treba da rezervise portove koji se koriste, da registruje uredjaj i da
 * alocira sve neophodne strukture
 */
{
	int errno;
	/*
	 * U vecini slucajeva, modul rezervise I/O portove koje koristi. Posto se 
	 * ovi portovi koriste za jos neke stvari, moraju biti deljeni medju
	 * modulima, a sinhronizacija se obezbedjuje kroz spinlock i8253_lock
	 * 
	 * U normalnom slucaju, pozvali bismo:
	 * if (request_region(ON_OFF_PORT, 0x01, "controll_register")==NULL) return -ENODEV;
	 * if (request_region(COUNTDOWN_UPDATE_PORT, 0x01, "tajmer")==NULL) return -ENODEV;
	*/
	
	/* alociramo strukture za sinhronizaciju */
	sema_init(&write_semafor, 1);
	sema_init(&ioctl_semafor, 1);
	sema_init(&read_semafor, 1);
	init_completion(&write_completion);
	complete(&write_completion);
	
	/* alociramo sebi jedan kod uredjaja */
	errno=alloc_chrdev_region(&telegraf_devnum, 0, DEVCOUNT, DEVNAME);
	if (errno!=0) goto chrdev_alloc_fail; else dev_ok=true;
	
	/* inicijalizujemo cdev strukturu sa konkretim operacijama za fajl */
	cdev_init(&telegraf_cdev, &telegraf_fops);
	/* povezujemo cdev strukturu i dobijeni ID */
	errno=cdev_add(&telegraf_cdev, telegraf_devnum, 1);
	
	if (errno!=0) goto cdev_alloc_fail; else cdev_ok=true;
	
	return 0;
	
	/* ako nesto ne uspe, dealociramo sve i vracamo kod greske */
	cdev_alloc_fail:
		cdev_del(&telegraf_cdev);
	chrdev_alloc_fail:
		complete(&write_completion);
		unregister_chrdev_region(telegraf_devnum, DEVCOUNT);
		return -ENODEV;
}
static void __exit pc_speaker_telegraf_exit(void)
/*
 * Funkcija koja se poziva pri izbacivanju modula iz kernela
 * dealocira sve sto smo alocirali
 */
{
	play_status=FORCE_STOP;
	wait_for_completion(&write_completion);
	if (cdev_ok) cdev_del(&telegraf_cdev);
	if (dev_ok) unregister_chrdev_region(telegraf_devnum, DEVCOUNT);
	//release_region(ON_OFF_PORT, 0x01);	// ovo preskacemo
}

static ssize_t telegraf_read(struct file *read_file, char __user *readbuffer, size_t read_size, loff_t *location)
/*
 * Funkcija koja se poziva kad neko uradi read nad fajlom /dev/telegraf
 * Posto read ovde nema mnogo smisla, samo se salje poruka da fajl ne treba
 * citati
 */
{
	// Kopiramo poruku u bafer response
	char *response;
	int data_remaining;
	down_interruptible(&read_semafor);
	response=vmalloc(sizeof(READ_RESPONSE)+1);
	if (response==NULL)
	{
		up(&read_semafor);
		return -ENOMEM;
	}
	strcpy(response, READ_RESPONSE);
	
	// kopiramo bafer korisniku
	if (read_size<=strlen(response))
	{
		data_remaining=copy_to_user(readbuffer, response, read_size);
		vfree(response);
		up(&read_semafor);
		if (data_remaining) return -EFAULT;
		else return read_size;
	}
	else
	{
		data_remaining=copy_to_user(readbuffer, response, strlen(response));
		vfree(response);
		up(&read_semafor);
		if (data_remaining) return -EFAULT;
		else return read_size;
	}
}
static ssize_t telegraf_write(struct file *write_file, const char __user *writebuffer, size_t write_size, loff_t *location)
/*
 * Funkcija koja obradjuje write operacije nad fajlom /dev/telegraf
 * Kopira upisani tekst u bafer i prosledjuje bafer niti koja emituje zvuk
 * na PC zvucniku
 */
{
	struct task_struct *playthread;
	char *towrite;
	int data_remaining;
	
	// alociramo bafer
	towrite=vmalloc(write_size+1);
	if (towrite==NULL)
	{
		// za slucaj da je alokacija neuspela
		printk(KERN_ERR "Failed to allocate memory for input from process %s (PID: %d),\
failed to allocate %lu bytes of memmory", current->comm, current->pid, write_size+1);
		return -ENOMEM;
	}
	
	// kopiramo iz userspace-a u bafer
	data_remaining=copy_from_user(towrite, writebuffer, write_size);
	if (data_remaining)
	{
		// u slucaju da kopiranje ne uspe
		vfree(towrite);
		return -EFAULT;
	}
	
	// dodajemo terminator na kraj stringa
	towrite[write_size-1]='\0';
	if (play_status==FORCE_STOP) 
	{
		vfree(towrite);
		return write_size;
	}
	
	// kreiramo nit, prosledjujemo joj string i startujemo je
	playthread=kthread_create(playstring, (void*)towrite, PLAY_THREAD_NAME);
	wake_up_process(playthread);	
	return write_size;
}

static int playstring(void *strw)
{
	size_t i;
	size_t wlen;
	char *str=(char*)strw;
	if (play_status==FORCE_STOP) 
	{
		vfree(str);
		return 0;
	}
	// zauzimamo semafor
	down_interruptible(&write_semafor);
	init_completion(&write_completion);	
	wlen=strlen(str);
	for (i=0;i<wlen;i++)
	{
		if (play_status==FORCE_STOP) 
		{
			// ako je neko u medjuvremenu zaustavio modul, stopiramo nit
			vfree(str);
			up(&write_semafor);
			complete_and_exit(&write_completion, 0);
		}
		// emitujemo kod za jedan karakter
		playchar(str[i]);
	}
	
	// brisemo bafer, oslobadjamo semafor i javljamo da je reprodukcija zvuka
	// gotova
	vfree(str);
	up(&write_semafor);
	complete_and_exit(&write_completion, 0);
}

static long telegraf_ioctl(struct file *telegraf_file, unsigned int ioctl_command, unsigned long ioctl_args)
/*
 * Funkcija koja obradjuje ioctl sistemske pozive
 */
{
	// proveravamo da li je korisnik admin (root)
	if ( !capable(CAP_SYS_ADMIN)) return -EPERM;
	
	// izvrsavanje komandi
	switch(ioctl_command)
	{
		case IOCTLCODE_SETFREQ:
		down_interruptible(&ioctl_semafor);
		freq=ioctl_args;
		countdown=TIMERFREQ/freq;
		countdown%=65536;
		highbyte=countdown>>8;
		lowbyte=countdown-highbyte;
		up(&ioctl_semafor);
		return 0;
		
		case IOCTLCODE_SETDOTLENGTH:
		dotlength=ioctl_args;
		return 0;
		
		default:	//neispravna komanda
		return -ENOTTY;		
	}
	return 0;
}

static void playchar(char letter)
/*
 * Reprodukuje zvukove za pojedinacan karakter po pravilima morzeove azbuke
 */
{
	if (letter>=65 && letter<=90) letter+=32;	// letter=tolower(letter)
	switch(letter)
	{
		case 'a':
			playdot();
			playdash();
			break;
		case 'b':
			playdash();
			playdot();
			playdot();
			playdot();
			break;
		case 'c':
			playdash();
			playdot();
			playdash();
			playdot();
			break;
		case 'd':
			playdash();
			playdot();
			playdot();
			break;
		case 'e':
			playdot();
			break;
		case 'f':
			playdot();
			playdot();
			playdash();
			playdot();
			break;
		case 'g':
			playdash();
			playdash();
			playdot();
			break;
		case 'h':
			playdot();
			playdot();
			playdot();
			playdot();
			break;
		case 'i':
			playdot();
			playdot();
			break;
		case 'j':
			playdot();
			playdash();
			playdash();
			playdash();
			break;
		case 'k':
			playdash();
			playdot();
			playdash();
			break;
		case 'l':
			playdot();
			playdash();
			playdot();
			playdot();;
			break;
		case 'm':
			playdash();
			playdash();
			break;
		case 'n':
			playdash();
			playdot();
			break;
		case 'o':
			playdash();
			playdash();
			playdash();
			break;
		case 'p':
			playdot();
			playdash();
			playdash();
			playdot();
			break;
		case 'q':
			playdash();
			playdash();
			playdot();
			playdash();
			break;
		case 'r':
			playdot();
			playdash();
			playdot();
			break;
		case 's':
			playdot();
			playdot();
			playdot();
			break;
		case 't':
			playdash();
			break;
		case 'u':
			playdot();
			playdot();
			playdash();
			break;
		case 'v':
			playdot();
			playdot();
			playdot();
			playdash();
			break;
		case 'w':
			playdot();
			playdash();
			playdash();
			break;
		case 'x':
			playdash();
			playdot();
			playdot();
			playdash();
			break;
		case 'y':
			playdash();
			playdot();
			playdash();
			playdash();
			break;
		case 'z':
			playdash();
			playdash();
			playdot();
			playdot();
			break;
		case '1':
			playdot();
			playdash();
			playdash();
			playdash();
			playdash();
			break;
		case '2':
			playdot();
			playdot();
			playdash();
			playdash();
			playdash();
			break;
		case '3':
			playdot();
			playdot();
			playdot();
			playdash();
			playdash();
			break;
		case '4':
			playdot();
			playdot();
			playdot();
			playdot();
			playdash();
			break;
		case '5':
			playdot();
			playdot();
			playdot();
			playdot();
			playdot();
		case '6':
			playdash();
			playdot();
			playdot();
			playdot();
			playdot();
			break;
		case '7':
			playdash();
			playdash();
			playdot();
			playdot();
			playdot();
			break;
		case '8':
			playdash();
			playdash();
			playdash();
			playdot();
			playdot();
			break;
		case '9':
			playdash();
			playdash();
			playdash();
			playdash();
			playdot();
			break;
		case '0':
			playdash();
			playdash();
			playdash();
			playdash();		
			playdash();
			break;
		case ' ':
			msleep_interruptible(dotlength*4);
			return;
		default:
			printk(KERN_NOTICE "Character '%c' not recognized\n", letter);
			return;
		}
		LETTERSLEEP;
	return;
}

static void playdot(void)
{
	unsigned char value;	
	unsigned long spinlock_flags;
	
	// javljamo tajmeru da cemo da azuriramo frekvenciju
	outb(CONTROLL_CODE_UPDATE_COUNTDOWN, PIT2_CONTROLL_PORT);
	smp_mb();
	
	// saljemo nizi bajt
	outb(lowbyte, COUNTDOWN_UPDATE_PORT);
	smp_mb();
	// saljemo visi bajt
	outb(highbyte, COUNTDOWN_UPDATE_PORT);
	smp_mb();
	// zakljucavamo pristup portu 0x61
	raw_spin_lock_irqsave(&i8253_lock, spinlock_flags);
	value=inb(ON_OFF_PORT);
	smp_mb();
	// palimo najniza 2 bita - startujemo tajmer i zvucnik
	value=value | 3;
	smp_mb();
	outb(value, ON_OFF_PORT);
	// otkljucavamo spinlock
	raw_spin_unlock_irqrestore(&i8253_lock, spinlock_flags);
	// cekamo da odsvira tacku
	msleep_interruptible(dotlength);
	
	// gasimo zvucnik i tajmer
	raw_spin_lock_irqsave(&i8253_lock, spinlock_flags);
	value=inb(ON_OFF_PORT);
	smp_mb();
	// gasimo najniza dva bita
	value=value & 252;
	smp_mb();
	outb(value, ON_OFF_PORT);
	raw_spin_unlock_irqrestore(&i8253_lock, spinlock_flags);
	DOTSLEEP;
}

static void playdash(void)
{
	unsigned char value;
	unsigned long spinlock_flags;
	outb(CONTROLL_CODE_UPDATE_COUNTDOWN, PIT2_CONTROLL_PORT);
	smp_mb();
	outb(lowbyte, COUNTDOWN_UPDATE_PORT);
	smp_mb();
	outb(highbyte, COUNTDOWN_UPDATE_PORT);
	raw_spin_lock_irqsave(&i8253_lock, spinlock_flags);
	value=inb(ON_OFF_PORT);
	smp_mb();
	value=value | 3;
	smp_mb();
	outb(value, ON_OFF_PORT);
	raw_spin_unlock_irqrestore(&i8253_lock, spinlock_flags);
	msleep_interruptible(dotlength*3);
	raw_spin_lock_irqsave(&i8253_lock, spinlock_flags);
	value=inb(ON_OFF_PORT);
	smp_mb();
	value=value & 252;
	smp_mb();
	outb(value, ON_OFF_PORT);
	raw_spin_unlock_irqrestore(&i8253_lock, spinlock_flags);
	DOTSLEEP;
}
/*
 * Druge zanimljive stvari:
 * 
 * izvoz simbola:
 * 		EXPORT_SYMBOL(name);
 * 		EXPORT_SYMBOL_GPL(name);
 * 
 * block uredjaji:
 * 		int register_blkdev(unsigned int major, const char *name);
 * 		struct gendisk *alloc_disk(int minors);
 * 		void request(request_queue_t *queue);
 * 
 * slab alokacija:
 * 		kmem_cache_t *kmem_cache_create(const char *name, size_t size, 
 * 										   size_t offset,
 *										   unsigned long flags,
 *										   void (*constructor)(void *, kmem_cache_t *, unsigned long flags),
 * 										   void (*destructor)(void *, kmem_cache_t *, unsigned long flags))
 * 		void *kmem_cache_alloc(kmem_cache_t *cache, int flags);
 * 		void kmem_cache_free(kmem_cache_t *cache, const void *obj);
 * 
 * memory mapped IO:
 * 		struct resource *request_mem_region(unsigned long start, unsigned long len, char *name);
 * 		void *ioremap(unsigned long phys_addr, unsigned long size);
 *		void *ioremap_nocache(unsigned long phys_addr, unsigned long size);
 *		unsigned int ioread8(void *addr);
 *		void iowrite8(u8 value, void *addr)
 * 
 * podrska za hotplug:
 * 		MODULE_DEVICE_TABLE(usb/pci, struct ...);
 * 		USB_DEVICE(vendor, product)
 */
